@@ -1,10 +1,10 @@
-# Copyrights 2008-2013 by [Mark Overmeer].
+# Copyrights 2008-2016 by [Mark Overmeer].
 #  For other contributors see ChangeLog.
 # See the manual pages for details on the licensing terms.
-# Pod stripped from pm file by OODoc 2.01.
+# Pod stripped from pm file by OODoc 2.02.
 package XML::LibXML::Simple;
 use vars '$VERSION';
-$VERSION = '0.93';
+$VERSION = '0.96';
 
 use base 'Exporter';
 
@@ -14,23 +14,25 @@ use warnings;
 our @EXPORT    = qw(XMLin);
 our @EXPORT_OK = qw(xml_in);
 
-use XML::LibXML    ();
-use File::Slurp    qw/read_file/;
-use File::Basename qw/fileparse/;
-use File::Spec     ();
+use XML::LibXML       ();
+use File::Slurp::Tiny qw/read_file/;
+use File::Basename    qw/fileparse/;
+use File::Spec        ();
 use Carp;
+use Scalar::Util      qw/blessed/;
 
 use Data::Dumper;  #to be removed
 
 
-my %known_opts = map { ($_ => 1) }
+my %known_opts = map +($_ => 1),
   qw(keyattr keeproot forcecontent contentkey noattr searchpath
      forcearray grouptags nsexpand normalisespace normalizespace
      valueattr nsstrip parser parseropts);
 
-my @DefKeyAttr     = qw(name key id);
-my $DefContentKey  = qq(content);
+my @default_attributes  = qw(name key id);
+my $default_content_key = 'content';
 
+#-------------
 
 sub new(@)
 {   my $class = shift;
@@ -44,23 +46,28 @@ sub new(@)
     $self;
 }
 
+#-------------
 
 sub XMLin
-{   my $self = @_ > 1 && UNIVERSAL::isa($_[0], __PACKAGE__) ? shift
+{   my $self = @_ > 1 && blessed $_[0] && $_[0]->isa(__PACKAGE__) ? shift
       : __PACKAGE__->new;
     my $target = shift;
 
-    my $this   = $self->_take_opts(@_);
-    my $opts   = $self->_init($self->{opts}, $this);
+    my $this = $self->_take_opts(@_);
+    my $opts = $self->_init($self->{opts}, $this);
 
-    my $xml    = $self->_get_xml($target, $opts)
+    my $xml  = $self->_get_xml($target, $opts)
         or return;
 
+    if(my $cb = $opts->{hooknodes})
+    {   $self->{XCS_hooks} = $cb->($self, $xml);
+    }
+
     my $top  = $self->collapse($xml, $opts);
     if($opts->{keeproot})
-    {    my $subtop
-           = $opts->{forcearray_always} && ref $top ne 'ARRAY' ? [$top] : $top;
-        $top = { $xml->localName => $subtop };
+    {   my $subtop
+          = $opts->{forcearray_always} && ref $top ne 'ARRAY' ? [$top] : $top;
+        $top = +{ $xml->localName => $subtop };
     }
 
     $top;
@@ -80,8 +87,9 @@ sub _get_xml($$)
               || $self->_create_parser($opts->{parseropts});
 
     my $xml
-      = UNIVERSAL::isa($source,'XML::LibXML::Document') ? $source
-      : UNIVERSAL::isa($source,'XML::LibXML::Element' ) ? $source
+      = blessed $source &&
+        (  $source->isa('XML::LibXML::Document')
+        || $source->isa('XML::LibXML::Element' )) ? $source
       : ref $source eq 'SCALAR' ? $parser->parse_string($$source)
       : ref $source             ? $parser->parse_fh($source)
       : $source =~ m{^\s*\<.*?\>\s*$}s ? $parser->parse_string($source)
@@ -149,7 +157,7 @@ sub _init($$)
 
     if(defined $opt{contentkey})
          { $opt{collapseagain} = $opt{contentkey} =~ s/^\-// }
-    else { $opt{contentkey} = $DefContentKey }
+    else { $opt{contentkey} = $default_content_key }
 
     $opt{normalisespace} ||= $opt{normalizespace} || 0;
 
@@ -172,7 +180,7 @@ sub _init($$)
     # Special cleanup for {keyattr} which could be arrayref or hashref,
     # which behave differently.
 
-    my $ka = $opt{keyattr} || \@DefKeyAttr;
+    my $ka = $opt{keyattr} || \@default_attributes;
     $ka    = [ $ka ] unless ref $ka;
  
     if(ref $ka eq 'ARRAY')
@@ -259,19 +267,32 @@ sub collapse($$)
     $xml->isa('XML::LibXML::Element') or return;
 
     my (%data, $text);
+    my $hooks = $self->{XCS_hooks};
 
     unless($opts->{noattr})
-    {   foreach my $attr ($xml->attributes)
-        {   my $value = $attr->value;
+    {
+      ATTR:
+        foreach my $attr ($xml->attributes)
+        {
+            my $value;
+            if($hooks && (my $hook = $hooks->{$attr->unique_key}))
+            {   $value = $hook->($attr);
+                defined $value or next ATTR;
+            }
+            else
+            {   $value = $attr->value;
+            }
+
             $value = $self->normalise_space($value)
-                if $opts->{normalisespace}==2;
+                if !ref $value && $opts->{normalisespace}==2;
 
-            my $n  = !$attr->isa('XML::LibXML::Attr') ? $attr->nodeName
-                   : $opts->{nsexpand}                ? _expand_name($attr)
-                   : $opts->{nsstrip}                 ? $attr->localName
-                   :                                    $attr->nodeName;
+            my $name
+              = !$attr->isa('XML::LibXML::Attr') ? $attr->nodeName
+              : $opts->{nsexpand} ? _expand_name($attr)
+              : $opts->{nsstrip}  ? $attr->localName
+              :                     $attr->nodeName;
 
-            _add_kv \%data, $n, $value, $opts;
+            _add_kv \%data, $name => $value, $opts;
         }
     }
     my $nr_attrs = keys %data;
@@ -280,17 +301,28 @@ sub collapse($$)
   CHILD:
     foreach my $child ($xml->childNodes)
     {
-        if($child->isa('XML::LibXML::Element'))
-        {   $nr_elems++;
-            my $v = $self->collapse($child, $opts);
-            my $n = $opts->{nsexpand} ? _expand_name($child)
-                  : $opts->{nsstrip}  ? $child->localName
-                  :                     $child->nodeName;
-            _add_kv \%data, $n, $v, $opts if defined $v;
-        }
-        elsif($child->isa('XML::LibXML::Text'))
+        if($child->isa('XML::LibXML::Text'))
         {   $text .= $child->data;
+            next CHILD;
         }
+
+        $child->isa('XML::LibXML::Element')
+            or next CHILD;
+
+        $nr_elems++;
+
+        my $v;
+        if($hooks && (my $hook = $hooks->{$child->unique_key}))
+             { $v = $hook->($child) }
+        else { $v = $self->collapse($child, $opts) }
+        defined $v or next CHILD;
+
+        my $name
+          = $opts->{nsexpand} ? _expand_name($child)
+          : $opts->{nsstrip}  ? $child->localName
+          :                     $child->nodeName;
+
+        _add_kv \%data, $name => $v, $opts;
     }
 
     $text = $self->normalise_space($text)
@@ -305,8 +337,8 @@ sub collapse($$)
     # Roll up 'value' attributes (but only if no nested elements)
 
     if(keys %data==1)
-    {    my ($k) = keys %data;
-         return $data{$k} if $opts->{valueattrlist}{$k};
+    {   my ($k) = keys %data;
+        return $data{$k} if $opts->{valueattrlist}{$k};
     }
 
     # Turn arrayrefs into hashrefs if key fields present
@@ -380,7 +412,7 @@ sub array_to_hash($$$$)
     my $ka = $opts->{keyattr} or return $in;
 
     if(ref $ka eq 'HASH')
-    {   my $newkey = $ka->{$name}  or return $in;
+    {   my $newkey = $ka->{$name} or return $in;
         my ($key, $flag) = @$newkey;
 
         foreach my $h (@$in)
@@ -408,7 +440,7 @@ sub array_to_hash($$$$)
     }
 
     else  # Arrayref
-    {   my $default_keys = "@DefKeyAttr" eq "@$ka";
+    {   my $default_keys = "@default_attributes" eq "@$ka";
 
       ELEMENT:
         foreach my $h (@$in)
@@ -454,7 +486,9 @@ sub array_to_hash($$$$)
 
     # first go through the values, checking that they are fit to collapse
     foreach my $v (values %out)
-    {   next if ref $v eq 'HASH' && keys %$v == 1 && $v->{$contentkey};
+    {   next if !defined $v;
+        next if ref $v eq 'HASH' && keys %$v == 1 && exists $v->{$contentkey};
+        next if ref $v eq 'HASH' && !keys %$v;
         return \%out;
     }
 
